==> rhgitaly-1.1.3dev0.tgz: rhgitaly-1.1.3dev0/rust/rhgitaly/src/repository.rs <==
// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::fmt;
use std::marker::Send;
use std::path::{Path, PathBuf};
use std::sync::Arc;

use tokio::fs;
use tokio::sync::mpsc;
use tokio_stream::wrappers::ReceiverStream;
use tonic::{Response, Status};

use hg::changelog::Changelog;
use hg::config::Config as CoreConfig;
use hg::repo::{Repo, RepoError};

use super::config::Config;
use super::gitaly::Repository;
use super::streaming::BlockingResponseSender;

/// Represent errors that are due to a wrong repository specification.
///
/// In terms of gRPC methods, the specifications is usually enclosed in  a [`Repository`] message
/// and these errors are considered to be client errors.
#[derive(Debug, PartialEq, Eq)]
pub enum RepoSpecError {
    MissingSpecification,
    UnknownStorage(String),
    RepoNotFound(PathBuf),
}

/// Represent errors loading a repository (bad specification or internal errors)
#[derive(Debug, derive_more::From)] // TODO add PartialEq, but do it in core for RepoError first
pub enum RepoLoadError {
    #[from]
    SpecError(RepoSpecError),
    LoadError(RepoError),
}

impl From<RepoError> for RepoLoadError {
    fn from(value: RepoError) -> Self {
        if let RepoError::NotFound { at } = value {
            return Self::SpecError(RepoSpecError::RepoNotFound(at));
        }
        Self::LoadError(value)
    }
}

/// Default conversion of ['RepoSpecError'] into a gRPC ['Status']
///
/// This function does not care to precisely match the error details, focusing on the error
/// codes instead.
///
/// The resulting codes match the most common behaviour of Gitaly, which actually behaves more
/// and more like this with time (e.g., as internal Git error get catched and handled).
pub fn default_repo_spec_error_status(err: RepoSpecError) -> Status {
    match err {
        RepoSpecError::MissingSpecification => Status::invalid_argument("repository not set"),
        RepoSpecError::UnknownStorage(storage) => Status::invalid_argument(format!(
            "GetStorageByName: no such storage: \"{}\"",
            storage
        )),
        RepoSpecError::RepoNotFound(at) => {
            Status::not_found(format!("repository at \"{}\" not found", at.display()))
        }
    }
}

pub fn repo_path(config: &Config, repo: &Repository) -> Result<PathBuf, RepoSpecError> {
    if repo.storage_name != "default" {
        return Err(RepoSpecError::UnknownStorage(repo.storage_name.clone()));
    }
    let root = &config.repositories_root;

    let res = match repo.relative_path.strip_suffix(".git") {
        Some(stripped) => root.join(&(stripped.to_owned() + ".hg")),
        None => root.join(&repo.relative_path),
    };

    // TODO forbid climbing up (same in Python, will be necessary for HGitaly3, even
    // though clients are deeply trusted.
    Ok(res)
}

/// Default gRPC error ['Status'] for repository not found.
///
/// To be used if repository path does not exist on disk.
pub fn default_repo_not_found_error_status(path: &Path) -> Status {
    Status::not_found(format!(
        "Mercurial repository at {} not found",
        path.display()
    ))
}

pub async fn checked_repo_path(
    config: &Config,
    gl_repo: &Option<Repository>,
) -> Result<PathBuf, RepoSpecError> {
    let repo = gl_repo
        .as_ref()
        .ok_or(RepoSpecError::MissingSpecification)?;
    let path = repo_path(config, repo)?;
    if match fs::metadata(&path).await {
        Ok(md) => md.is_dir(),
        Err(_) => false,
    } {
        return Ok(path);
    }
    Err(RepoSpecError::RepoNotFound(path))
}

/// Return a path to virtual filesystem for the repository store.
///
/// As of this writing, this is nothing but a [`Path`], but it could become something
/// more advanced in the future (perhaps not as much as `hg_core` `Vfs` type, though).
///
/// Parameter `repo` is an `Option`, so that a service method can pass directly
/// something like `&request.repository`, with `None` giving rise to the natural
/// `MissingSpecification` error.
///
/// If the repository is not found on disc, the appropriate error is also returned.
pub async fn repo_store_vfs(
    config: &Config,
    repo: &Option<Repository>,
) -> Result<PathBuf, RepoSpecError> {
    let root = checked_repo_path(config, repo).await?;
    Ok(root.join(".hg/store"))
}

pub fn load_repo(config: &Config, opt_repo: Option<&Repository>) -> Result<Repo, RepoLoadError> {
    let repo = opt_repo.ok_or(RepoSpecError::MissingSpecification)?;
    // TODO load core config once and for all and store it in our Config.
    // but do we really want to have all this unrelated stuff around?
    let core_config =
        CoreConfig::load_non_repo().expect("Should have been able to read Mercurial config");
    // TODO better to use Repo::new_at_path, but it is private
    // (Repo::find does verifications that are useless to us.
    // At least it does not try to climb up when passed an explicit path)
    Ok(Repo::find(&core_config, Some(repo_path(config, repo)?))?)
}

/// Trait for requests with a repository field
///
/// It provides the needed uniformity for methods such as [`load_repo_and_stream`]
pub trait RequestWithRepo: Send + 'static {
    /// Grab a reference to the [`Repository`] field from the request.
    ///
    /// Like all submessages, the repository is optional.
    fn repository_ref(&self) -> Option<&Repository>;
}

/// Load a repository and initiate streaming responses
///
/// This setups the `mpsc` channel expected by Tonic and spawns a blocking task (typically run
/// in a separate thread) loads the repository, and passes over the repository and the transmitting
/// end of the channel to the caller supplied closure.
///
/// The `and_then` closure must perform its streaming by sending `Result<Resp, Status>` values
/// on the channel, using the provided [`BlockingResponseSender`].
///
/// If the repository loading fails, the appropriate gRPC error response is sent over
/// or logged if sending is not possible.
///
/// Because Gitaly's error responses are not uniform, and we want to match them closely,
/// ethe caller has to supply a callable for conversion of [`RepoSpecError`] to the appropriate
/// [`Status`]. The [`default_repo_spec_error_status`] function can be used in the majority of
/// cases and serve as an example for other cases.
pub fn load_repo_and_stream<Req: RequestWithRepo, Resp: fmt::Debug + Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, Repo, BlockingResponseSender<Resp>) + Send + 'static,
) -> Result<Response<ReceiverStream<Result<Resp, Status>>>, Status> {
    // no point having channel capacity for several messages, since `blocking_send` is the only
    // way to use it.
    let (tx, rx) = mpsc::channel(1);
    tokio::task::spawn_blocking(move || {
        let btx: BlockingResponseSender<Resp> = tx.into();
        match load_repo(&config, request.repository_ref()) {
            Err(RepoLoadError::SpecError(e)) => btx.send(Err(repo_spec_error_status(e))),
            Err(RepoLoadError::LoadError(e)) => btx.send(Err(Status::internal(format!(
                "Error loading repository: {:?}",
                e
            )))),
            Ok(repo) => and_then(request, repo, btx),
        }
    });
    Ok(Response::new(ReceiverStream::new(rx)))
}

/// Load a repository in a separate thread and hand it over to a closure
///
/// This creates a new thread suitable for blocking operations, using
/// [`tokio::task::spawn_blocking`] and then hands it over together with the orignal request
/// to the `and_then` closure, whose return value is finally returned to the caller.
/// The closure is at liberty to use any blocking operation (most calls to `hg-core` are blocking).
///
/// If the repository loading fails, the appropriate gRPC error [`Status`] is returned
///
/// `repo_spec_error_status` plays the same role as in [`load_repo_and_stream`], and again
/// the [`default_repo_spec_error_status`] function can be used in the majority of
/// cases.
pub async fn load_repo_and_then<Req: RequestWithRepo, Res: Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, Repo) -> Result<Res, Status> + Send + 'static,
) -> Result<Res, Status> {
    tokio::task::spawn_blocking(move || match load_repo(&config, request.repository_ref()) {
        Err(RepoLoadError::SpecError(e)) => Err(repo_spec_error_status(e)),
        Err(RepoLoadError::LoadError(e)) => Err(Status::internal(format!(
            "Error loading repository: {:?}",
            e
        ))),
        Ok(repo) => and_then(request, repo),
    })
    .await
    .map_err(|e| Status::internal(format!("Unexpected error in inner thread: {:?}", e)))?
}

/// Load a repository and its changelog in a separate thread and hand it over to a closure
///
/// See [`load_repo_and_then`] on which this builds upon for more details.
pub async fn load_changelog_and_then<Req: RequestWithRepo, Res: Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, &Repo, &Changelog) -> Result<Res, Status> + Send + 'static,
) -> Result<Res, Status> {
    load_repo_and_then(config, request, repo_spec_error_status, |req, repo| {
        let cl = repo
            .changelog()
            .map_err(|e| Status::internal(format!("Could not open changelog: {:?}", e)))?;
        and_then(req, &repo, &cl)
    })
    .await
}

/// Load a repository and its changelog in a separate thread for streaming responses
///
/// See [`load_repo_and_stream`] on which this builds upon for more details.
pub fn load_changelog_and_stream<Req: RequestWithRepo, Resp: fmt::Debug + Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, &Repo, &Changelog, BlockingResponseSender<Resp>) + Send + 'static,
) -> Result<Response<ReceiverStream<Result<Resp, Status>>>, Status> {
    load_repo_and_stream(
        config,
        request,
        repo_spec_error_status,
        |req, repo, tx| match repo.changelog() {
            Ok(cl) => and_then(req, &repo, &cl, tx),
            Err(e) => tx.send(Err(Status::internal(format!(
                "Could not open changelog: {:?}",
                e
            )))),
        },
    )
}

#[cfg(test)]
mod tests {

    use super::*;

    #[test]
    fn test_repo_path() {
        let config = Config {
            repositories_root: "/repos".into(),
            listen_url: "".into(),
        };
        let mut repo = Repository::default();
        repo.storage_name = "default".into();
        repo.relative_path = "foo/bar".into();
        assert_eq!(repo_path(&config, &repo), Ok("/repos/foo/bar".into()));

        repo.relative_path = "foo/bar.git".into();
        assert_eq!(repo_path(&config, &repo), Ok("/repos/foo/bar.hg".into()));
    }
}

==> rhgitaly-1.4.1.tgz: rhgitaly-1.4.1/rust/rhgitaly/src/repository.rs <==
// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::fmt;
use std::marker::Send;
use std::path::{Path, PathBuf};
use std::sync::Arc;

use tokio::fs;
use tokio::sync::mpsc;
use tokio_stream::wrappers::ReceiverStream;
use tonic::{Response, Status};

use hg::changelog::Changelog;
use hg::config::Config as CoreConfig;
use hg::repo::{Repo, RepoError};

use super::config::Config;
use super::gitaly::Repository;
use super::streaming::BlockingResponseSender;

/// Represent errors that are due to a wrong repository specification.
///
/// In terms of gRPC methods, the specifications is usually enclosed in  a [`Repository`] message
/// and these errors are considered to be client errors.
#[derive(Debug, PartialEq, Eq)]
pub enum RepoSpecError {
    MissingSpecification,
    UnknownStorage(String),
    RepoNotFound(PathBuf),
}

/// Represent errors loading a repository (bad specification or internal errors)
#[derive(Debug, derive_more::From)] // TODO add PartialEq, but do it in core for RepoError first
pub enum RepoLoadError {
    #[from]
    SpecError(RepoSpecError),
    LoadError(RepoError),
}

impl From<RepoError> for RepoLoadError {
    fn from(value: RepoError) -> Self {
        if let RepoError::NotFound { at } = value {
            return Self::SpecError(RepoSpecError::RepoNotFound(at));
        }
        Self::LoadError(value)
    }
}

/// Default conversion of ['RepoSpecError'] into a gRPC ['Status']
///
/// This function does not care to precisely match the error details, focusing on the error
/// codes instead.
///
/// The resulting codes match the most common behaviour of Gitaly, which actually behaves more
/// and more like this with time (e.g., as internal Git error get catched and handled).
pub fn default_repo_spec_error_status(err: RepoSpecError) -> Status {
    match err {
        RepoSpecError::MissingSpecification => Status::invalid_argument("repository not set"),
        RepoSpecError::UnknownStorage(storage) => Status::invalid_argument(format!(
            "GetStorageByName: no such storage: \"{}\"",
            storage
        )),
        RepoSpecError::RepoNotFound(at) => {
            Status::not_found(format!("repository at \"{}\" not found", at.display()))
        }
    }
}

pub fn repo_path(config: &Config, repo: &Repository) -> Result<PathBuf, RepoSpecError> {
    if repo.storage_name != "default" {
        return Err(RepoSpecError::UnknownStorage(repo.storage_name.clone()));
    }
    let root = &config.repositories_root;

    let res = match repo.relative_path.strip_suffix(".git") {
        Some(stripped) => root.join(stripped.to_owned() + ".hg"),
        None => root.join(&repo.relative_path),
    };

    // TODO forbid climbing up (same in Python, will be necessary for HGitaly3, even
    // though clients are deeply trusted.
    Ok(res)
}

/// Default gRPC error ['Status'] for repository not found.
///
/// To be used if repository path does not exist on disk.
pub fn default_repo_not_found_error_status(path: &Path) -> Status {
    Status::not_found(format!(
        "Mercurial repository at {} not found",
        path.display()
    ))
}

pub async fn checked_repo_path(
    config: &Config,
    gl_repo: &Option<Repository>,
) -> Result<PathBuf, RepoSpecError> {
    let repo = gl_repo
        .as_ref()
        .ok_or(RepoSpecError::MissingSpecification)?;
    let path = repo_path(config, repo)?;
    if match fs::metadata(&path).await {
        Ok(md) => md.is_dir(),
        Err(_) => false,
    } {
        return Ok(path);
    }
    Err(RepoSpecError::RepoNotFound(path))
}

/// Return a path to virtual filesystem for the repository store.
///
/// As of this writing, this is nothing but a [`Path`], but it could become something
/// more advanced in the future (perhaps not as much as `hg_core` `Vfs` type, though).
///
/// Parameter `repo` is an `Option`, so that a service method can pass directly
/// something like `&request.repository`, with `None` giving rise to the natural
/// `MissingSpecification` error.
///
/// If the repository is not found on disc, the appropriate error is also returned.
pub async fn repo_store_vfs(
    config: &Config,
    repo: &Option<Repository>,
) -> Result<PathBuf, RepoSpecError> {
    let root = checked_repo_path(config, repo).await?;
    Ok(root.join(".hg/store"))
}

pub fn load_repo(config: &Config, opt_repo: Option<&Repository>) -> Result<Repo, RepoLoadError> {
    let repo = opt_repo.ok_or(RepoSpecError::MissingSpecification)?;
    // TODO load core config once and for all and store it in our Config.
    // but do we really want to have all this unrelated stuff around?
    let core_config =
        CoreConfig::load_non_repo().expect("Should have been able to read Mercurial config");
    // TODO better to use Repo::new_at_path, but it is private
    // (Repo::find does verifications that are useless to us.
    // At least it does not try to climb up when passed an explicit path)
    Ok(Repo::find(&core_config, Some(repo_path(config, repo)?))?)
}

/// Trait for requests with a repository field
///
/// It provides the needed uniformity for methods such as [`load_repo_and_stream`]
pub trait RequestWithRepo: Send + 'static {
    /// Grab a reference to the [`Repository`] field from the request.
    ///
    /// Like all submessages, the repository is optional.
    fn repository_ref(&self) -> Option<&Repository>;
}

/// Load a repository and initiate streaming responses
///
/// This setups the `mpsc` channel expected by Tonic and spawns a blocking task (typically run
/// in a separate thread) loads the repository, and passes over the repository and the transmitting
/// end of the channel to the caller supplied closure.
///
/// The `and_then` closure must perform its streaming by sending `Result<Resp, Status>` values
/// on the channel, using the provided [`BlockingResponseSender`].
///
/// If the repository loading fails, the appropriate gRPC error response is sent over
/// or logged if sending is not possible.
///
/// Because Gitaly's error responses are not uniform, and we want to match them closely,
/// ethe caller has to supply a callable for conversion of [`RepoSpecError`] to the appropriate
/// [`Status`]. The [`default_repo_spec_error_status`] function can be used in the majority of
/// cases and serve as an example for other cases.
pub fn load_repo_and_stream<Req: RequestWithRepo, Resp: fmt::Debug + Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, Repo, BlockingResponseSender<Resp>) + Send + 'static,
) -> Result<Response<ReceiverStream<Result<Resp, Status>>>, Status> {
    // no point having channel capacity for several messages, since `blocking_send` is the only
    // way to use it.
    let (tx, rx) = mpsc::channel(1);
    tokio::task::spawn_blocking(move || {
        let btx: BlockingResponseSender<Resp> = tx.into();
        match load_repo(&config, request.repository_ref()) {
            Err(RepoLoadError::SpecError(e)) => btx.send(Err(repo_spec_error_status(e))),
            Err(RepoLoadError::LoadError(e)) => btx.send(Err(Status::internal(format!(
                "Error loading repository: {:?}",
                e
            )))),
            Ok(repo) => and_then(request, repo, btx),
        }
    });
    Ok(Response::new(ReceiverStream::new(rx)))
}

/// Load a repository in a separate thread and hand it over to a closure
///
/// This creates a new thread suitable for blocking operations, using
/// [`tokio::task::spawn_blocking`] and then hands it over together with the orignal request
/// to the `and_then` closure, whose return value is finally returned to the caller.
/// The closure is at liberty to use any blocking operation (most calls to `hg-core` are blocking).
///
/// If the repository loading fails, the appropriate gRPC error [`Status`] is returned
///
/// `repo_spec_error_status` plays the same role as in [`load_repo_and_stream`], and again
/// the [`default_repo_spec_error_status`] function can be used in the majority of
/// cases.
pub async fn load_repo_and_then<Req: RequestWithRepo, Res: Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, Repo) -> Result<Res, Status> + Send + 'static,
) -> Result<Res, Status> {
    tokio::task::spawn_blocking(move || match load_repo(&config, request.repository_ref()) {
        Err(RepoLoadError::SpecError(e)) => Err(repo_spec_error_status(e)),
        Err(RepoLoadError::LoadError(e)) => Err(Status::internal(format!(
            "Error loading repository: {:?}",
            e
        ))),
        Ok(repo) => and_then(request, repo),
    })
    .await
    .map_err(|e| Status::internal(format!("Unexpected error in inner thread: {:?}", e)))?
}

/// Load a repository and its changelog in a separate thread and hand it over to a closure
///
/// See [`load_repo_and_then`] on which this builds upon for more details.
pub async fn load_changelog_and_then<Req: RequestWithRepo, Res: Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, &Repo, &Changelog) -> Result<Res, Status> + Send + 'static,
) -> Result<Res, Status> {
    load_repo_and_then(config, request, repo_spec_error_status, |req, repo| {
        let cl = repo
            .changelog()
            .map_err(|e| Status::internal(format!("Could not open changelog: {:?}", e)))?;
        and_then(req, &repo, &cl)
    })
    .await
}

/// Load a repository and its changelog in a separate thread for streaming responses
///
/// See [`load_repo_and_stream`] on which this builds upon for more details.
pub fn load_changelog_and_stream<Req: RequestWithRepo, Resp: fmt::Debug + Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, &Repo, &Changelog, BlockingResponseSender<Resp>) + Send + 'static,
) -> Result<Response<ReceiverStream<Result<Resp, Status>>>, Status> {
    load_repo_and_stream(
        config,
        request,
        repo_spec_error_status,
        |req, repo, tx| match repo.changelog() {
            Ok(cl) => and_then(req, &repo, &cl, tx),
            Err(e) => tx.send(Err(Status::internal(format!(
                "Could not open changelog: {:?}",
                e
            )))),
        },
    )
}

#[cfg(test)]
mod tests {

    use super::*;

    #[test]
    fn test_repo_path() {
        let config = Config {
            repositories_root: "/repos".into(),
            listen_url: "".into(),
        };
        let mut repo = Repository::default();
        repo.storage_name = "default".into();
        repo.relative_path = "foo/bar".into();
        assert_eq!(repo_path(&config, &repo), Ok("/repos/foo/bar".into()));

        repo.relative_path = "foo/bar.git".into();
        assert_eq!(repo_path(&config, &repo), Ok("/repos/foo/bar.hg".into()));
    }
}
